### einfache-schematische-tiefensuche-2018.scm                    
from ist_Euler_mit_gallenbacher_graph_kanten import *

### Loesung basiert auf:
### ===== schematische-Tiefensuche ==================================
### Die Tiefensuche ist darin abstrakt definiert. Erst die
### jeweilige Implementation der Funktion Nachfolger und die
### Behandlung der Alternativen als Liste haengt von der Art
### der Repraesentation des Graphen ab.
### Es wird hier davon ausgegangen, dass dieser in einer
### Kantenliste vorliegt.
### Muster :
### [ (1.Knoten, 2.Knoten, Bewertung), ... ]
###
### Die folgende Variante ermoeglicht nur eine Suche nach der
### ersten Loesung.
### ===== Tiefensuche ===============================================
# fuehrt eine Tiefensuche mit backtracking unter Verwendung
# des Systemstacks und einer besucht-Liste (besuchteKanten) aus.
# Die Ausgabe des Weges erfolgt aus dieser Liste, wenn
# eine Suche erfolgreich war.
# Durch die interne Funktion wird das Übergeben von Start, Ziel
# und Graph vermieden.

### ----- speziell Eulerscher Weg -----
# Hier muss das Ziel gleich dem Start sein, daher muss ein Erfolg in
# jedem Fall vor dem Pruefen eines Zyklus untersucht und ausgegeben werden.

def Tiefensuche(Start, Graph):

    def tS(alternativen, besuchteKanten): # bei Euler die Kanten
        ##input((alternativen,besuchteKanten))
        if len(alternativen)==0:                          # [Anm.1]
            return False
        
        if len(besuchteKanten)==anzahl:
            if besuchteKanten[0][1]==Start:               # [Anm.2]
                return besuchteKanten          # Loesung
            return False                       # keine Loesung

        if kanteVorhanden(alternativen[0], besuchteKanten):  # [Anm.3]
            return tS(alternativen[1:], besuchteKanten)
                                             
        ergebnis=tS(nachfolgeKanten(alternativen[0][1], Graph),   # [Anm.4]:
                    [alternativen[0]]+besuchteKanten)
        if ergebnis:
            return ergebnis
        return tS(alternativen[1:], besuchteKanten)         # [Anm.5]
    return tS(nachfolgeKanten(Start, Graph), [])

# [Anm.1]: Wenn auf einer Rekursionstiefe alle Alternativen
#   untersucht wurden, wird mit dem Wert "#f" zur aufrufenden
#   Ebene zurueckgesprungen.
# [Anm.2]: Wenn das Ziel erreicht wurde, wird die Kantenliste
#   zur aufrufenden Ebene zurueckgegeben.
# [Anm.3]: Es ist ein Zyklus aufgetreten, weiter suchen.
#   Hier muss aber auch nach der Gegenkante gesucht werden!
# [Anm.4]: Hier erfolgt der eigentliche Abstieg in der Rekursion.
#   Die aktuelle Kante wird in  besuchteKanten aufgenommen.
# [Anm.5]: Wenn die Rekursion in der Tiefe erfolglos war, muss noch
#   die Rekursion fuer die Alternativen versucht werden.

### ===== nachfolgeKanten ================================================
### Die Funktion nachfolgeKanten ist zentral notwendig für die Loesung.
### Sie generiert jeweils zum aktuellen Knoten die Liste aller
### Nachfolgekanten.
# ===== Hilfsfunktion istNachfolgeKante =====
def istNachfolgeKante(aktuelleStadt, kante):
    return aktuelleStadt==kante[0]

# ===== nachfolgeKanten =====
# Muster fuer die Funktion Nachfolger
def nachfolgeKanten(aktuelleStadt, kanten):
    nachF=[]
    for kante in kanten:
        if istNachfolgeKante(aktuelleStadt, kante):
            nachF.append(kante)
    return nachF
# Test:
#(nachfolgeKanten 'I (kanten-raus '(I C 40) (kanten-raus '(O P 91) *kanten*)))
#(nachfolgeKanten 'G  *kanten*)

### ----- kanteVorhanden ------------------------------------------------
# prueft, ob die Kante bereits besucht wurde
def kanteVorhanden(kante, besuchteKanten):
    ##print(kante, besuchteKanten)
    if (kante[1],kante[0],kante[2]) in besuchteKanten:
        return True # ist Rueckkante
    return kante in besuchteKanten

### ===== Laufzeit ==================================================
import time
Start='I'
anzahl = len(eulerGraph) // 2
start=time.perf_counter()
ts=Tiefensuche(Start, eulerGraph)
stopp=time.perf_counter()
print(ts,'\nZeit:',(stopp-start)*1000,'ms\n')

